import {
  API,
  DynamicPlatformPlugin,
  Logger,
  PlatformAccessory,
  PlatformConfig,
  Service,
  Characteristic,
} from "homebridge";
import { IConfig, isConfig } from "./config";
import * as faceapi from "@vladmandic/face-api";
import canvas from "canvas";
import fs from "fs";
import * as path from "path";
import { nets } from "@vladmandic/face-api";
import { FaceMatcher } from "@vladmandic/face-api";
import { Monitor } from "./monitor/monitor";
import { Trainer } from "./trainer";
require("@tensorflow/tfjs-node");

const { Canvas, Image, ImageData } = canvas;
//@ts-ignore
faceapi.env.monkeyPatch({ Canvas, Image, ImageData });

import { PLATFORM_NAME, PLUGIN_NAME } from "./settings";
import { LocationAccessory } from "./locationAccessory";

/**
 * HomebridgePlatform
 * This class is the main constructor for your plugin, this is where you should
 * parse the user config and discover/register accessories with Homebridge.
 */
export class HomeLocationPlatform implements DynamicPlatformPlugin {
  public readonly Service: typeof Service = this.api.hap.Service;
  public readonly Characteristic: typeof Characteristic = this.api.hap
    .Characteristic;

  // this is used to track restored cached accessories
  public readonly accessories: PlatformAccessory[] = [];
  public config: IConfig;

  constructor(
    public readonly log: Logger,
    config: PlatformConfig,
    public readonly api: API
  ) {
    this.log.debug("Finished initializing platform:", config.name);

    if (!isConfig(config)) {
      this.log.error("Configuration is incorrect or incomplete");
      process.exit(1);
    } else {
      this.config = config;
    }

    this.api.on("didFinishLaunching", async () => {
      log.debug("Executed didFinishLaunching callback");
      // run the method to discover / register your devices as accessories
      await this.discoverDevices();
    });
  }

  /**
   * This function is invoked when homebridge restores cached accessories from disk at startup.
   * It should be used to setup event handlers for characteristics and update respective values.
   */
  public configureAccessory(accessory: PlatformAccessory) {
    this.log.info("Loading accessory from cache:", accessory.displayName);

    // add the restored accessory to the accessories cache so we can track if it has already been registered
    this.accessories.push(accessory);
  }

  /**
   * This is an example method showing how to register discovered accessories.
   * Accessories must only be registered once, previously created accessories
   * must not be registered again to prevent "duplicate UUID" errors.
   */
  public async discoverDevices() {
    const faceDetectionNet = nets.ssdMobilenetv1;
    await faceDetectionNet.loadFromDisk(this.config.weightDirectory);
    await nets.faceLandmark68Net.loadFromDisk(this.config.weightDirectory);
    await nets.faceRecognitionNet.loadFromDisk(this.config.weightDirectory);

    //Train facial recognition model
    let faceMatcher: FaceMatcher;
    if (this.config.trainOnStartup) {
      const trainer = new Trainer(
        this.config.refImageDirectory,
        this.config.trainedModelDirectory,
        this.config.confidence
      );
      faceMatcher = await trainer.train(true);
    } else {
      const raw = fs.readFileSync(
        path.join(this.config.trainedModelDirectory, "data.json"),
        "utf-8"
      );
      faceMatcher = FaceMatcher.fromJSON(JSON.parse(raw));
    }

    const locationMonitor = new Monitor(
      this.config.rooms,
      faceMatcher,
      this.log,
      this.config
    );

    locationMonitor.startStreams();

    const labels = faceMatcher.labeledDescriptors.map((e) => e.label);
    for (const room of this.config.rooms) {
      const uuid = this.api.hap.uuid.generate(room.name);

      const existingAccessory = this.accessories.find((e) => e.UUID === uuid);
      if (existingAccessory) {
        this.log.info(
          "Restoring existing accessory from cache: ",
          existingAccessory.displayName
        );

        new LocationAccessory(this, existingAccessory, locationMonitor, room);

        this.api.updatePlatformAccessories([existingAccessory]);
      } else {
        this.log.info("Adding new accessory:", `${room.name}`);

        // create a new accessory
        const accessory = new this.api.platformAccessory(`${room.name}`, uuid);

        accessory.context["DeviceName"] = `${room.name}`;

        // create the accessory handler for the newly create accessory
        // this is imported from `platformAccessory.ts`
        new LocationAccessory(this, accessory, locationMonitor, room);

        // link the accessory to your platform
        this.api.registerPlatformAccessories(PLUGIN_NAME, PLATFORM_NAME, [
          accessory,
        ]);
      }
    }
  }
}
